use macros::decode_args;
use num_derive::FromPrimitive;
use num_traits::FromPrimitive;
use truth::llir::RawInstr;

#[derive(Debug, FromPrimitive)]
enum Opcode {
  Halt = 0,
  Jump = 1,
  Position = 2,
  PositionInterp = 3,
  Facing = 4,
  FacingInterp = 5,
  Rotation = 6,
  FieldOfView = 7,
  Fog = 8,
  FogInterp = 9,
  PositionInterpBezier = 10,
  FacingInterpBezier = 11,
  ShakingMode = 12,
  BackgroundFillColor = 13,
  BackgroundSprite = 14,
}

#[derive(Debug)]
pub enum Op {
  Halt,
  Jump {
    index: usize,
    time: i32,
  },
  Position {
    x: f32,
    y: f32,
    z: f32,
  },
  PositionInterp {
    time: i32,
    mode: i32,
    x: f32,
    y: f32,
    z: f32,
  },
  Facing {
    x: f32,
    y: f32,
    z: f32,
  },
  FacingInterp {
    time: i32,
    mode: i32,
    x: f32,
    y: f32,
    z: f32,
  },
  Rotation {
    x: f32,
    y: f32,
    z: f32,
  },
  FieldOfView(f32),
  Fog {
    color: i32,
    start: f32,
    end: f32,
  },
  FogInterp {
    time: i32,
    mode: i32,
    color: i32,
    start: f32,
    end: f32,
  },
  ShakingMode {
    mode: i32,
  },
}

pub struct Instruction {
  pub time: i32,
  pub op: Op,
}

impl Instruction {
  pub fn from_raw(inst: RawInstr, instruction_offsets: &Vec<(u32, usize)>) -> Self {
    let offset_to_index =
      |value: i32| instruction_offsets.binary_search_by_key(&(value as u32), |a| a.0).expect("failed to find entry");

    let args = &inst.args_blob;
    let opcode_raw = inst.opcode;
    let opcode = Opcode::from_u16(opcode_raw).expect("failed to convert opcode");

    let op = match opcode {
      Opcode::Halt => Op::Halt,
      Opcode::Jump => {
        let (offset, time) = decode_args!(args, "ot");
        Op::Jump {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::Position => {
        let (x, y, z) = decode_args!(args, "fff");
        Op::Position { x, y, z }
      }
      Opcode::PositionInterp => {
        let (time, mode, x, y, z) = decode_args!(args, "SUfff");
        Op::PositionInterp { time, mode, x, y, z }
      }
      Opcode::Facing => {
        let (x, y, z) = decode_args!(args, "fff");
        Op::Facing { x, y, z }
      }
      Opcode::FacingInterp => {
        let (time, mode, x, y, z) = decode_args!(args, "SUfff");
        Op::FacingInterp { time, mode, x, y, z }
      }
      Opcode::Rotation => {
        let (x, y, z) = decode_args!(args, "fff");
        Op::Rotation { x, y, z }
      }
      Opcode::FieldOfView => {
        let fov = decode_args!(args, "f");
        Op::FieldOfView(fov)
      }
      Opcode::Fog => {
        let (color, start, end) = decode_args!(args, "Sff");
        Op::Fog { color, start, end }
      }
      Opcode::FogInterp => {
        let (time, mode, color, start, end) = decode_args!(args, "SUSff");
        Op::FogInterp {
          time,
          mode,
          color,
          start,
          end,
        }
      }
      Opcode::PositionInterpBezier => todo!(),
      Opcode::FacingInterpBezier => todo!(),
      Opcode::ShakingMode => todo!(),
      Opcode::BackgroundFillColor => todo!(),
      Opcode::BackgroundSprite => todo!(),
    };

    Instruction { time: inst.time, op }
  }
}
