use proc_macro2::{Span, TokenStream};
use quote::quote;
use syn::{
  parse::{ParseStream, Parser},
  spanned::Spanned,
  Expr, ExprLit, Ident, Lit, LitStr, Token, Type, TypeArray,
};
use truth::{
  context::RootEmitter,
  llir::{ArgEncoding, InstrAbi, StringArgSize},
  pos::SourceStr,
};

#[proc_macro]
pub fn decode_args(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
  match blob_impl.parse(input) {
    Ok(data) => data.into(),
    Err(error) => error.into_compile_error().into(),
  }
}

fn blob_impl(input: ParseStream) -> syn::Result<TokenStream> {
  let reader: Expr = input.parse()?;
  let _: Token![,] = input.parse()?;
  let sig_str: LitStr = input.parse()?;
  let sig_str_span = sig_str.span();
  let sig_str = sig_str.value();
  let (sig_str, variadic) = if sig_str.ends_with('v') {
    (&sig_str.as_str()[..sig_str.len() - 1], true)
  } else {
    (sig_str.as_str(), false)
  };
  let sig = InstrAbi::parse(SourceStr::from_full_source(None, sig_str), &RootEmitter::new_stderr()).unwrap();
  let encodings: Vec<_> = sig
    .arg_encodings()
    .map(|enc| {
      Ok(match enc {
        ArgEncoding::Float => quote!(cursor.read_f32().unwrap()),
        ArgEncoding::Integer { size, .. } => {
          let size = *size as usize;

          match size {
            2 => quote!(cursor.read_i16().unwrap()),
            4 => quote!(cursor.read_i32().unwrap()),
            _ => panic!("invalid integer size {size}"),
          }
        }
        ArgEncoding::JumpTime => quote!(cursor.read_i32().unwrap()),
        ArgEncoding::JumpOffset => quote!(cursor.read_i32().unwrap()),
        ArgEncoding::Color => quote!(cursor.read_u32().unwrap()),
        ArgEncoding::String {
          size: StringArgSize::Fixed { len, .. },
          ..
        } => quote!(cursor.read_cstring_exact(#len).unwrap().decode(truth::io::DEFAULT_ENCODING).unwrap()),
        ArgEncoding::String {
          size: StringArgSize::Pascal { block_size },
          ..
        } => quote!(cursor.read_cstring_blockwise(#block_size).unwrap().decode(truth::io::DEFAULT_ENCODING).unwrap()),
        _ => return Err(syn::Error::new(sig_str_span, "failed...")),
      })
    })
    .collect::<syn::Result<_>>()?;

  let variadic = variadic.then(|| {
    quote! {
      , {
        let file_size = cursor.file_size().unwrap();
        let offset = (file_size - cursor.pos().unwrap()) as usize;
        cursor.read_byte_vec(offset).unwrap()
      }
    }
  });

  Ok(quote! {
    {
      use truth::io::BinRead;
      let mut cursor = ::std::io::Cursor::new(#reader);
      (
        #(#encodings),*
        #variadic
      )
    }
  })
}

//spawner_rank_decode_args!(args, "S", [[f32; 2]; 3])

#[proc_macro]
pub fn spawner_param_decode_args(input: proc_macro::TokenStream) -> proc_macro::TokenStream {
  match spawner_param_decode_args_impl.parse(input) {
    Ok(data) => data.into(),
    Err(error) => error.into_compile_error().into(),
  }
}

fn spawner_param_decode_args_impl(input: ParseStream) -> syn::Result<TokenStream> {
  let reader: Expr = input.parse()?;
  let _: Token![,] = input.parse()?;
  let sig_str: LitStr = input.parse()?;
  let _: Token![,] = input.parse()?;
  let array = input.parse::<TypeArray>()?;

  let Expr::Lit(ExprLit { lit: Lit::Int(int), .. }) = array.len else {
    return Err(syn::Error::new(array.len.span(), "length must be a constant number"));
  };
  let rank_count = int.base10_parse::<usize>()?;
  let Type::Array(TypeArray {
    elem,
    len: Expr::Lit(ExprLit { lit: Lit::Int(int), .. }),
    ..
  }) = array.elem.as_ref()
  else {
    return Err(syn::Error::new(
      array.elem.span(),
      "element was not an array with constant number ",
    ));
  };
  let entry_count = int.base10_parse::<usize>()?;
  let Type::Path(path) = elem.as_ref() else {
    return Err(syn::Error::new(elem.span(), "expected path as element"));
  };
  let ident = path.path.get_ident().ok_or(syn::Error::new(path.span(), "path was not a single ident"))?;
  let type_id = ident.to_string();
  enum ParamType {
    Float,
    Int,
  }
  let param_ty = match type_id.as_str() {
    "f32" => ParamType::Float,
    "i32" => ParamType::Int,
    _ => return Err(syn::Error::new(ident.span(), "expected i32 or f32")),
  };

  let total = entry_count * rank_count;
  let mut arguments = sig_str.value();

  arguments.extend(std::iter::repeat_n(
    match param_ty {
      ParamType::Int => 'S',
      ParamType::Float => 'f',
    },
    total,
  ));

  let rank_list: Vec<Vec<Ident>> = (0..rank_count)
    .map(|rank| {
      (0..entry_count).map(move |entry| Ident::new(&format!("__a_{rank}_{entry}"), Span::call_site())).collect()
    })
    .collect();

  let tuple_pat = rank_list.iter().flatten();
  let rank_array = rank_list.iter().enumerate().map(|(rank_index, list)| {
    let list = list.iter().enumerate().map(|(entry_index, entry)| {
      let param_index = rank_index * entry_count + entry_index + 1;
      let param_index = param_index as u16;
      let kind = match param_ty {
        ParamType::Float => quote!(param_float),
        ParamType::Int => quote!(param_int),
      };
      quote!(#kind(#entry, #param_index))
    });
    quote! {
      [#(#list),*]
    }
  });

  Ok(quote! {
    {
      let (id, #(#tuple_pat),*) = decode_args!(#reader, #arguments);
      (id, [#(#rank_array),*])
    }
  })
}
