use std::{collections::VecDeque, num::NonZero, sync::Weak};

use bytemuck::{bytes_of, Pod, Zeroable};
use glam::{Mat4, Vec2, Vec3, Vec4};
use num_traits::FloatConst;
use wgpu::{
  include_wgsl,
  naga::FastHashMap,
  util::{BufferInitDescriptor, DeviceExt},
  BindGroupDescriptor, BindGroupEntry, BindGroupLayoutDescriptor, BindGroupLayoutEntry, BindingResource, BlendState,
  BufferDescriptor, BufferUsages, Color, ColorTargetState, ColorWrites, DepthStencilState, Device, Extent3d,
  FragmentState, LoadOp, Operations, PipelineLayoutDescriptor, PrimitiveState, RenderPass, RenderPassColorAttachment,
  RenderPassDepthStencilAttachment, RenderPassDescriptor, RenderPipeline, RenderPipelineDescriptor, ShaderStages,
  StoreOp, TextureDescriptor, TextureFormat, TextureUsages, VertexAttribute, VertexBufferLayout, VertexState,
};
use winit::dpi::PhysicalSize;

use crate::{
  engine::{Engine, UpdateContext},
  game::anm::{loaded_file::SpriteUvs, vm::RenderingState, AnmVm},
};

use super::{Manager, Uniform, WeakVm};

#[derive(Default, Clone, Copy, Pod, Zeroable)]
#[repr(C, packed)]
pub(super) struct Instance {
  matrix: Mat4,
  sprite_uv: SpriteUvs,
  color: Vec4,
  anchor_offset: Vec2,
}

impl Manager {
  pub fn new(context: &UpdateContext) -> Manager {
    let depth_texture = Self::create_depth_texture(context.device, context.window.inner_size());

    let texture_bind_group_layout = context.device.create_bind_group_layout(&BindGroupLayoutDescriptor {
      entries: &[BindGroupLayoutEntry {
        binding: 0,
        ty: wgpu::BindingType::Texture {
          sample_type: wgpu::TextureSampleType::Float { filterable: true },
          view_dimension: wgpu::TextureViewDimension::D2,
          multisampled: false,
        },
        count: None,
        visibility: ShaderStages::FRAGMENT,
      }],
      label: None,
    });

    let bind_group_layout = context.device.create_bind_group_layout(&BindGroupLayoutDescriptor {
      label: None,
      entries: &[
        BindGroupLayoutEntry {
          binding: 0,
          ty: wgpu::BindingType::Sampler(wgpu::SamplerBindingType::Filtering),
          count: None,
          visibility: ShaderStages::FRAGMENT,
        },
        BindGroupLayoutEntry {
          binding: 1,
          count: None,
          visibility: ShaderStages::VERTEX,
          ty: wgpu::BindingType::Buffer {
            ty: wgpu::BufferBindingType::Uniform,
            has_dynamic_offset: false,
            min_binding_size: Some(NonZero::new(std::mem::size_of::<Uniform>() as u64).unwrap()),
          },
        },
      ],
    });

    let sampler = context.device.create_sampler(&wgpu::SamplerDescriptor {
      label: Some("sampler"),
      address_mode_u: wgpu::AddressMode::ClampToEdge,
      address_mode_v: wgpu::AddressMode::ClampToEdge,
      address_mode_w: wgpu::AddressMode::ClampToEdge,
      mag_filter: wgpu::FilterMode::Linear,
      min_filter: wgpu::FilterMode::Nearest,
      mipmap_filter: wgpu::FilterMode::Nearest,
      ..Default::default()
    });

    let ui_uniform = Uniform {
      proj_matrix: Mat4::orthographic_lh(0.0, 640.0, 480.0, 0.0, 100.0, -100.0),
    };

    let uniform_buffer = context.device.create_buffer_init(&BufferInitDescriptor {
      contents: bytes_of(&ui_uniform),
      label: Some("uniform"),
      usage: BufferUsages::UNIFORM | BufferUsages::COPY_DST,
    });

    let render_bind_group = context.device.create_bind_group(&BindGroupDescriptor {
      label: Some("render bind group"),
      layout: &bind_group_layout,
      entries: &[
        BindGroupEntry {
          binding: 0,
          resource: BindingResource::Sampler(&sampler),
        },
        BindGroupEntry {
          binding: 1,
          resource: BindingResource::Buffer(uniform_buffer.as_entire_buffer_binding()),
        },
      ],
    });

    let pipeline_layout = context.device.create_pipeline_layout(&PipelineLayoutDescriptor {
      label: None,
      bind_group_layouts: &[&texture_bind_group_layout, &bind_group_layout],
      push_constant_ranges: &[],
    });

    let render_shader = context.device.create_shader_module(include_wgsl!("../render.wgsl"));
    let blit_shader = context.device.create_shader_module(include_wgsl!("../blit_screen.wgsl"));
    let clear_shader = context.device.create_shader_module(include_wgsl!("../clear.wgsl"));

    let clear_pipeline = context.device.create_render_pipeline(&RenderPipelineDescriptor {
      label: Some("clear"),
      layout: Some(&context.device.create_pipeline_layout(&PipelineLayoutDescriptor {
        label: None,
        bind_group_layouts: &[],
        push_constant_ranges: &[],
      })),
      vertex: VertexState {
        buffers: &[],
        module: &clear_shader,
        compilation_options: Default::default(),
        entry_point: "vs_main",
      },
      primitive: PrimitiveState {
        topology: wgpu::PrimitiveTopology::TriangleStrip,
        ..Default::default()
      },
      depth_stencil: Some(DepthStencilState {
        format: TextureFormat::Depth32Float,
        depth_write_enabled: true,
        depth_compare: wgpu::CompareFunction::Less,
        stencil: wgpu::StencilState::default(),
        bias: wgpu::DepthBiasState::default(),
      }),
      multisample: Default::default(),
      fragment: Some(FragmentState {
        targets: &[Some(ColorTargetState {
          format: context.config.format,
          blend: Some(BlendState::REPLACE),
          write_mask: ColorWrites::ALL,
        })],
        compilation_options: Default::default(),
        entry_point: "fs_main",
        module: &clear_shader,
      }),
      multiview: None,
      cache: None,
    });

    let blit_pipeline = context.device.create_render_pipeline(&RenderPipelineDescriptor {
      label: Some("blit"),
      layout: Some(&pipeline_layout),
      vertex: VertexState {
        buffers: &[],
        module: &blit_shader,
        compilation_options: Default::default(),
        entry_point: "vs_main",
      },
      primitive: PrimitiveState {
        topology: wgpu::PrimitiveTopology::TriangleStrip,
        ..Default::default()
      },
      depth_stencil: Some(DepthStencilState {
        format: TextureFormat::Depth32Float,
        depth_write_enabled: true,
        depth_compare: wgpu::CompareFunction::Less,
        stencil: wgpu::StencilState::default(),
        bias: wgpu::DepthBiasState::default(),
      }),
      multisample: Default::default(),
      fragment: Some(FragmentState {
        targets: &[Some(ColorTargetState {
          format: context.config.format,
          blend: Some(BlendState::ALPHA_BLENDING),
          write_mask: ColorWrites::ALL,
        })],
        compilation_options: Default::default(),
        entry_point: "fs_main",
        module: &blit_shader,
      }),
      multiview: None,
      cache: None,
    });
    let world_uniform = Uniform {
      proj_matrix: Mat4::perspective_lh(
        45.0 * (f32::PI() / 180.0),
        context.config.width as f32 / context.config.height as f32,
        0.1,
        1000.0,
      ),
    };

    let (sender, receiver) = crossbeam::channel::bounded(2);
    Manager {
      anm_files: FastHashMap::default(),
      anm_sender: sender,
      anm_receiver: receiver,

      world_backbuffer_anm: Weak::new(),
      ui_vms: VecDeque::new(),
      world_vms: VecDeque::new(),

      depth_texture,
      ui_uniform,
      world_uniform,
      uniform_buffer,
      render_bind_group,
      render_pipeline_layout: pipeline_layout,
      render_shader,
      clear_pipeline,
      blit_pipeline,
    }
  }

  fn new_pipeline(&self, engine: &Engine, vm: &AnmVm) -> RenderPipeline {
    let render_pipeline = engine.device.create_render_pipeline(&RenderPipelineDescriptor {
      label: Some("anm"),
      layout: Some(&self.render_pipeline_layout),
      vertex: VertexState {
        buffers: &[VertexBufferLayout {
          attributes: &[
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x4,
              offset: 16 * 0,
              shader_location: 0,
            },
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x4,
              offset: 16 * 1,
              shader_location: 1,
            },
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x4,
              offset: 16 * 2,
              shader_location: 2,
            },
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x4,
              offset: 16 * 3,
              shader_location: 3,
            },
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x4,
              offset: 16 * 4,
              shader_location: 4,
            },
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x4,
              offset: 16 * 5,
              shader_location: 5,
            },
            VertexAttribute {
              format: wgpu::VertexFormat::Float32x2,
              offset: 16 * 6,
              shader_location: 6,
            },
          ],
          array_stride: std::mem::size_of::<Instance>() as u64,
          step_mode: wgpu::VertexStepMode::Instance,
        }],
        entry_point: "vs_main",
        compilation_options: Default::default(),
        module: &self.render_shader,
      },
      primitive: PrimitiveState {
        topology: wgpu::PrimitiveTopology::TriangleStrip,
        ..Default::default()
      },
      depth_stencil: Some(DepthStencilState {
        format: TextureFormat::Depth32Float,
        depth_write_enabled: false,
        depth_compare: wgpu::CompareFunction::Less,
        stencil: wgpu::StencilState::default(),
        bias: wgpu::DepthBiasState::default(),
      }),
      multisample: Default::default(),
      fragment: Some(FragmentState {
        targets: &[Some(ColorTargetState {
          format: engine.config.format,
          blend: Some(vm.blend_state),
          write_mask: ColorWrites::ALL,
        })],
        compilation_options: Default::default(),
        entry_point: "fs_main",
        module: &self.render_shader,
      }),
      multiview: Default::default(),
      cache: Default::default(),
    });

    render_pipeline
  }

  fn render_layer(&self, engine: &Engine, pass: &mut RenderPass, list: &VecDeque<WeakVm>, layer: u32) {
    for vm in list {
      if let Some(vm) = vm.upgrade() {
        let vm = vm.borrow();
        if vm.layer != layer {
          self.render_vm(engine, pass, &vm, &Vec3::new(0.0, 0.0, layer as f32));
        }
      }
    }
  }

  fn render_vm(&self, engine: &Engine, pass: &mut RenderPass, vm: &AnmVm, layer_origin: &Vec3) {
    if !vm.visible {
      return;
    }

    if let Some((sprite, entry)) = vm.sprite_entry_pair() {
      let mut rendering_state = vm.rendering_state.borrow_mut();
      if rendering_state.is_none() {
        let instance_buffer = engine.device.create_buffer(&BufferDescriptor {
          label: Some("instances"),
          usage: BufferUsages::VERTEX | BufferUsages::COPY_DST,
          mapped_at_creation: false,
          size: std::mem::size_of::<Instance>() as u64,
        });

        *rendering_state = Some(RenderingState {
          instance_buffer: instance_buffer,
          pipeline: Some(self.new_pipeline(engine, vm)),
        });
      }

      let rendering_state = rendering_state.as_mut().unwrap();

      if rendering_state.pipeline.is_none() {
        rendering_state.pipeline = Some(self.new_pipeline(engine, vm))
      }

      let mut matrix = Mat4::IDENTITY;
      matrix *= Mat4::from_translation(layer_origin + vm.position());
      matrix *= Mat4::from_rotation_translation(vm.rotation(), Vec3::ZERO);
      matrix *= Mat4::from_scale((sprite.size * vm.scale()).extend(1.0));

      engine.queue.write_buffer(
        &rendering_state.instance_buffer,
        0,
        bytes_of(&Instance {
          matrix,
          sprite_uv: sprite.uvs + vm.scroll(),
          color: vm.color().extend(vm.alpha()),
          anchor_offset: vm.anchor_offset,
        }),
      );

      pass.set_pipeline(rendering_state.pipeline.as_ref().unwrap());
      pass.set_bind_group(0, &entry.bind_group, &[]);
      pass.set_bind_group(1, &self.render_bind_group, &[]);
      pass.set_vertex_buffer(0, rendering_state.instance_buffer.slice(..));
      pass.draw(0..4, 0..1 as u32);
    }

    for child in vm.children.iter() {
      let child = child.borrow();
      self.render_vm(engine, pass, &child, &layer_origin);
    }
  }

  fn render_ui(&self, engine: &Engine, encoder: &mut wgpu::CommandEncoder, surface: &wgpu::Texture) {
    engine.queue.write_buffer(&self.uniform_buffer, 0, bytes_of(&self.ui_uniform));
    let mut pass = encoder.begin_render_pass(&RenderPassDescriptor {
      label: Some("ui anms"),
      color_attachments: &[Some(RenderPassColorAttachment {
        view: &surface.create_view(&Default::default()),
        resolve_target: None,
        ops: Operations {
          // load: LoadOp::Clear(Color {
          //   // r: 100. / 255.,
          //   // g: 149. / 255.,
          //   // b: 237. / 255.,
          //   r: 1.0,
          //   g: 1.0,
          //   b: 1.0,
          //   a: 1.0,
          // }),
          load: LoadOp::Load,
          ..Default::default()
        },
      })],
      depth_stencil_attachment: Some(RenderPassDepthStencilAttachment {
        view: &self.depth_texture.create_view(&Default::default()),
        depth_ops: Some(Operations {
          load: LoadOp::Clear(1.0),
          store: StoreOp::Store,
        }),
        stencil_ops: None,
      }),
      ..Default::default()
    });

    self.render_layer(engine, &mut pass, &self.ui_vms, 21);
    self.render_layer(engine, &mut pass, &self.ui_vms, 22);
    self.render_layer(engine, &mut pass, &self.ui_vms, 23);
    self.render_layer(engine, &mut pass, &self.ui_vms, 29);
  }

  fn render_world(&self, engine: &Engine, encoder: &mut wgpu::CommandEncoder, surface: &wgpu::Texture) {
    engine.queue.write_buffer(&self.uniform_buffer, 0, bytes_of(&self.world_uniform));
    let mut pass = encoder.begin_render_pass(&RenderPassDescriptor {
      label: Some("world anms"),
      color_attachments: &[Some(RenderPassColorAttachment {
        view: &surface.create_view(&Default::default()),
        resolve_target: None,
        ops: Operations {
          // load: LoadOp::Clear(Color {
          //   // r: 100. / 255.,
          //   // g: 149. / 255.,
          //   // b: 237. / 255.,
          //   r: 1.0,
          //   g: 1.0,
          //   b: 1.0,
          //   a: 1.0,
          // }),
          load: LoadOp::Load,
          ..Default::default()
        },
      })],
      depth_stencil_attachment: Some(RenderPassDepthStencilAttachment {
        view: &self.depth_texture.create_view(&Default::default()),
        depth_ops: Some(Operations {
          load: LoadOp::Clear(1.0),
          store: StoreOp::Store,
        }),
        stencil_ops: None,
      }),
      ..Default::default()
    });

    self.render_layer(engine, &mut pass, &self.world_vms, 0);
    self.render_layer(engine, &mut pass, &self.world_vms, 1);
    self.render_layer(engine, &mut pass, &self.world_vms, 2);
    self.render_layer(engine, &mut pass, &self.world_vms, 3);
    self.render_layer(engine, &mut pass, &self.world_vms, 8);

  }

  fn clear(&self, encoder: &mut wgpu::CommandEncoder, surface: &wgpu::Texture) {
    let mut pass = encoder.begin_render_pass(&RenderPassDescriptor {
      label: Some("clear"),
      color_attachments: &[Some(RenderPassColorAttachment {
        view: &surface.create_view(&Default::default()),
        resolve_target: None,
        ops: Operations {
          load: LoadOp::Clear(Color {
            r: 100. / 255.,
            g: 149. / 255.,
            b: 0. / 255.,
            a: 1.0,
          }),
          ..Default::default()
        },
      })],
      depth_stencil_attachment: Some(RenderPassDepthStencilAttachment {
        view: &self.depth_texture.create_view(&Default::default()),
        depth_ops: Some(Operations {
          load: LoadOp::Clear(1.0),
          store: StoreOp::Store,
        }),
        stencil_ops: None,
      }),
      ..Default::default()
    });

    pass.set_pipeline(&self.clear_pipeline);
    pass.draw(0..4, 0..1);
  }

  pub fn render(&self, engine: &Engine, surface: &wgpu::Texture, encoder: &mut wgpu::CommandEncoder) {
    self.clear(encoder, surface);
    self.render_world(engine, encoder, surface);
    self.render_ui(engine, encoder, surface);

    if let Some(backbuffer_anm) = self.world_backbuffer_anm.upgrade() {
      let mut pass = encoder.begin_render_pass(&RenderPassDescriptor {
        label: Some("blit"),
        color_attachments: &[Some(RenderPassColorAttachment {
          view: &surface.create_view(&Default::default()),
          resolve_target: None,
          ops: Operations {
            load: LoadOp::Clear(Color {
              r: 100. / 255.,
              g: 149. / 255.,
              b: 0. / 255.,
              a: 1.0,
            }),
            ..Default::default()
          },
        })],
        depth_stencil_attachment: Some(RenderPassDepthStencilAttachment {
          view: &self.depth_texture.create_view(&Default::default()),
          depth_ops: Some(Operations {
            load: LoadOp::Load,
            store: StoreOp::Store,
          }),
          stencil_ops: None,
        }),
        ..Default::default()
      });

      pass.set_pipeline(&self.blit_pipeline);
      pass.set_bind_group(
        0,
        &backbuffer_anm.borrow().sprite_entry_pair().unwrap().1.bind_group,
        &[],
      );
      pass.set_bind_group(1, &self.render_bind_group, &[]);
      pass.draw(0..4, 0..1);
    }
  }

  pub(super) fn create_depth_texture(device: &Device, size: PhysicalSize<u32>) -> wgpu::Texture {
    device.create_texture(&TextureDescriptor {
      label: Some("depth texture"),
      size: Extent3d {
        width: size.width.max(1),
        height: size.height.max(1),
        depth_or_array_layers: 1,
      },
      mip_level_count: 1,
      sample_count: 1,
      dimension: wgpu::TextureDimension::D2,
      format: wgpu::TextureFormat::Depth32Float,
      usage: TextureUsages::RENDER_ATTACHMENT | TextureUsages::TEXTURE_BINDING,
      view_formats: &[],
    })
  }

  pub fn resize(&mut self, device: &Device, new_size: PhysicalSize<u32>) {
    self.depth_texture = Self::create_depth_texture(device, new_size)
  }
}
