use std::sync::Arc;

use anm::LoadedFile;
use states::GameStateMachine;

use crate::{
  engine::{EngineState, UpdateContext},
  utils::{game::Game, soon::Soon},
};

mod anm;
mod stage;
mod snd;
mod states;
mod timer;

pub struct GameRunner {
  sound_manager: snd::Manager,
  anm_manager: anm::Manager,
  state: GameStateMachine,
}

struct GameContext<'a> {
  engine: &'a UpdateContext<'a>,
  sound_manager: &'a mut snd::Manager,
  anm_manager: &'a mut anm::Manager,
}

impl GameContext<'_> {
  pub fn start_load_anm(&mut self, game: Game, file_name: impl Into<String>) -> Soon<Arc<LoadedFile>> {
    self.anm_manager.start_load_anm(
      self.engine.device,
      self.engine.queue,
      self.engine.window.inner_size(),
      game,
      file_name,
    )
  }

  pub fn load_anm(&mut self, game: Game, file_name: impl Into<String>) -> Arc<LoadedFile> {
    self.anm_manager.load_anm(
      self.engine.device,
      self.engine.queue,
      self.engine.window,
      game,
      file_name,
    )
  }
}

impl EngineState for GameRunner {
  fn new(context: &UpdateContext) -> Self
  where
    Self: Sized,
  {
    let mut sound_manager = snd::Manager::new();
    let mut anm_manager = anm::Manager::new(&context);

    let mut context = GameContext {
      anm_manager: &mut anm_manager,
      sound_manager: &mut sound_manager,
      engine: context,
    };

    let state = GameStateMachine::new(&mut context);

    let game = Self {
      sound_manager,
      anm_manager,
      state,
    };

    game
  }

  fn resize(&mut self, context: &UpdateContext, new_size: winit::dpi::PhysicalSize<u32>) {
    self.anm_manager.resize(context.device, new_size);
  }

  fn update(&mut self, _context: &UpdateContext) {
    self.state.update(&mut GameContext {
      engine: _context,
      sound_manager: &mut self.sound_manager,
      anm_manager: &mut self.anm_manager,
    });
    self.anm_manager.update();
  }

  fn render(&self, engine: &crate::engine::Engine, surface: &wgpu::Texture, encoder: &mut wgpu::CommandEncoder) {
    self.anm_manager.render(&engine, surface, encoder)
  }
}
