use std::io::Cursor;

use async_std::fs;
use instructions::std::Instruction;
use truth::{context::RootEmitter, io::BinReader, std::StdExtra, Game as TruthGame, StdFile};
use wgpu::naga::FastHashMap;

use crate::{
  game::{
    anm::{self, Vm, VmLocation},
    GameContext,
  },
  utils::game::Game,
};

pub struct LoadedFile {
  pub layers: FastHashMap<u16, Vec<Vm>>,
  pub instructions: Vec<instructions::std::Instruction>,
}

impl LoadedFile {
  pub async fn load(context: &mut GameContext<'_>, game: Game, file_name: impl Into<String> + Send) -> LoadedFile {
    let file_name = file_name.into();
    let file_data = fs::read(game.asset_path(&file_name)).await.expect("failed to load anm file");
    let file = StdFile::read_from_stream(
      &mut BinReader::from_reader(&RootEmitter::new_stderr(), &file_name, Cursor::new(file_data)),
      match game {
        Game::Mof => TruthGame::Th10,
        Game::Sa => TruthGame::Th11,
      },
    )
    .unwrap();
    let StdExtra::Th10 { anm_path } = file.extra else {
      unreachable!()
    };
    let anm_file = context.start_load_anm(game, anm_path.value).await;

    let mut layers: FastHashMap<u16, Vec<Vm>> = FastHashMap::default();

    for instance in file.instances {
      let object = file.objects.get(&instance.object).expect("failed to get object");
      for quad in &object.quads {
        let vm = context.anm_manager.new_vm(anm_file.clone(), None, quad.anm_script.into(), VmLocation::new());
        layers.entry(object.layer).or_default().push(vm);
      }
    }
    let mut offset_instructions: Vec<(_, _)> = Default::default();
    let mut current_offset = 0;
    for (index, inst) in file.script.iter().enumerate() {
      offset_instructions.push((current_offset, index));
      current_offset += 8 + inst.args_blob.len() as u32
    }

    let instructions =
      file.script.into_iter().map(|inst| Instruction::from_raw(inst, &offset_instructions)).collect::<Vec<_>>();
    LoadedFile { layers, instructions }
  }
}
