use sfsm::*;
use winit::keyboard::KeyCode;

use crate::{
  game::{
    anm::{Vm, VmLocation},
    states::UPDATE_CONTEXT,
  },
  utils::game::Game,
};

pub struct MainMenu(MainMenuMachine);
impl MainMenu {
  pub fn new() -> MainMenu {
    UPDATE_CONTEXT.with(|context| {
      let title = context.load_anm(Game::Mof, "title.anm");
      let title_ver = context.load_anm(Game::Mof, "title_v.anm");

      let mut main_menu = MainMenu(MainMenuMachine::new());
      main_menu
        .0
        .start(Boot {
          logo: context.anm_manager.new_vm(title, None, 90, VmLocation::new_ui()),
          title_ver: context.anm_manager.new_vm(title_ver, None, 0, VmLocation::new_ui()),
        })
        .unwrap();

      main_menu
    })
  }

  pub fn switch_to_gameplay(&self) -> bool {
    sfsm::IsState::<StartingGameplay>::is_state(&self.0)
  }
}
impl State for MainMenu {
  fn execute(&mut self) {
    self.0.step().unwrap();
  }
}

add_state_machine!(MainMenuMachine, Boot, {Boot, Starting, Selecting, StartingGameplay, Exiting}, {
  Boot => Starting,
  Starting => Selecting,
  Selecting => StartingGameplay,
  Selecting => Exiting,
});
struct Boot {
  logo: Vm,
  title_ver: Vm,
}
impl State for Boot {
  fn execute(&mut self) {
    self.exit();
    UPDATE_CONTEXT.with(|context| {
      if context.engine.keys.is_any_key_down() {
        self.logo.borrow_mut().interrupt(6);
        self.title_ver.borrow_mut().interrupt(1);
      }
    });
  }
}

impl Transition<Starting> for Boot {
  fn guard(&self) -> TransitGuard {
    self.title_ver.borrow_mut().deleted().into()
  }
}
impl From<Boot> for Starting {
  fn from(value: Boot) -> Self {
    Starting {
      logo: value.logo,
      time_in_state: 0,
    }
  }
}

struct Starting {
  logo: Vm,
  time_in_state: u32,
}
impl State for Starting {
  fn execute(&mut self) {
    self.time_in_state += 1;
  }
}

impl Transition<Selecting> for Starting {
  fn guard(&self) -> TransitGuard {
    (self.time_in_state >= 30).into()
  }
}
impl From<Starting> for Selecting {
  fn from(val: Starting) -> Self {
    UPDATE_CONTEXT.with(|context| {
      let file = val.logo.borrow().file().clone();
      Selecting {
        logo: val.logo,
        main_ui: context.anm_manager.new_vm(file, None, 0, VmLocation::new_child_ui(0)),
        selected: 0,
        selection_changed: false,
        input_locked: false,
      }
    })
  }
}

struct Selecting {
  logo: Vm,
  main_ui: Vm,
  selected: u32,
  selection_changed: bool,
  input_locked: bool,
}
impl Selecting {
  fn selected(&self, desired: u32) -> TransitGuard {
    (self.selected == desired && self.input_locked).into()
  }
}
impl State for Selecting {
  fn execute(&mut self) {
    UPDATE_CONTEXT.with(|context| {
      if self.input_locked {
        return;
      }
      if context.engine.keys.was_key_pressed(KeyCode::ArrowUp) {
        if let Some(prev) = self.selected.checked_sub(1) {
          self.selected = prev;
        } else {
          self.selected = 7;
        }
        if self.selected == 1 {
          self.selected = 0;
        }

        self.selection_changed = false;
      }

      if context.engine.keys.was_key_pressed(KeyCode::ArrowDown) {
        if self.selected >= 7 {
          self.selected = 0;
        } else {
          self.selected += 1;
        }
        if self.selected == 1 {
          self.selected = 2;
        }

        self.selection_changed = false;
      }

      if context.engine.keys.was_key_pressed(KeyCode::KeyZ) {
        self.main_ui.borrow_mut().interrupt(3);
        self.input_locked = true;
      }

      if !self.selection_changed {
        self.selection_changed = true;
        let main_ui = self.main_ui.borrow_mut();
        for i in 0..8u32 {
          let child = main_ui.child(i as usize).unwrap();
          let child_bg = main_ui.child(8 + i as usize).unwrap();
          if i == self.selected {
            child.borrow_mut().interrupt(i + 7);
            child_bg.borrow_mut().interrupt(i + 7);
          } else {
            context.anm_manager.interrupt_immediately(&child, 3);
            context.anm_manager.interrupt_immediately(&child_bg, 3);
          }
        }
      }
    });
  }
}

#[derive(Default)]
struct StartingGameplay;
derive_state!(StartingGameplay);
derive_transition_into_default!(Selecting, StartingGameplay);
impl Transition<StartingGameplay> for Selecting {
  fn guard(&self) -> TransitGuard {
    self.selected(0)
  }

  fn action(&mut self) {
    println!("agmlng")
  }
}

impl Transition<Exiting> for Selecting {
  fn guard(&self) -> TransitGuard {
    self.selected(7)
  }
}
impl From<Selecting> for Exiting {
  fn from(value: Selecting) -> Self {
    Exiting {
      _logo: value.logo,
      _main_ui: value.main_ui,
      timer: 0,
    }
  }
}

struct Exiting {
  _logo: Vm,
  _main_ui: Vm,
  timer: u32,
}

impl State for Exiting {
  fn execute(&mut self) {
    self.timer += 1;
    if self.timer >= 20 {
      UPDATE_CONTEXT.with(|context| context.engine.should_exit.set(true));
    }
  }
}
